/*
	ush -- minimalistic shell
	(c) 2006  Avelino Herrera Morales

	* no command completion - no wild chars - no pipes
	* environment variable expansion using traditional $VARIABLE_NAME syntax
	* builtin commands:
		cd DIRECTORY
		export VARIABLE=value
		exit
	* any other command is execlp'ed
	* stdin, stdout and stderr redirection using traditional ">", ">>", "2>", "2>>" and "<" sequences


	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>


char input[1024];
char expanded[1024];
char complete_path[1024];
char **complete_argv = NULL;
int child_pid;
void (*prev_signal_handler)(int);
FILE *input_file;
char output_redirect_file[1024];
int output_redirect_append = 0;
char error_redirect_file[1024];
int error_redirect_append = 0;
char input_redirect_file[1024];

/*
	expand the environment variables at "in" to their values
*/
char *expand(char *in) {
	char *p = NULL;
	char *remain_p = in;
	char *expanded_p = expanded;
	int variable_name_size;
	char *variable_name;
	char *variable_value;
	int variable_value_size;

	while ((p = strchr(remain_p, '$')) != NULL) {
		while (remain_p < p) {
			*expanded_p = *remain_p;
			expanded_p++;
			remain_p++;
		}
		variable_name_size = strspn(p + 1, "abdcefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_");
		/*printf("DEBUG: variable_name_size=%d\n", variable_name_size);*/
		variable_name = malloc(variable_name_size + 1);
		remain_p += variable_name_size + 1;
		variable_value = NULL;
		variable_value_size;
		memcpy(variable_name, p + 1, variable_name_size);
		variable_name[variable_name_size] = 0;
		/*printf("DEBUG: variable_name='%s'\n", variable_name);*/
		variable_value = getenv(variable_name);
		free(variable_name);
		if (variable_value == NULL)
			continue;
		variable_value_size = strlen(variable_value);
		memcpy(expanded_p, variable_value, variable_value_size);
		expanded_p += variable_value_size;
	}
	while (*remain_p != 0) {
		*expanded_p = *remain_p;
		expanded_p++;
		remain_p++;
	}
	*expanded_p = 0;
	return expanded;
}


#define  REDIRECT_TYPE_INPUT   0
#define  REDIRECT_TYPE_OUTPUT  1
#define  REDIRECT_TYPE_ERROR   2
/*
	get the name of the redirect file at the command line (NULL if no redirect specified)
*/
char *get_redirect_name(char *command_line, int redirect_type, char *dest, int *is_append) {
	char *handle_p;
	char *start;
	char *output_p;
	char *first;
	char *d;
	if (redirect_type == REDIRECT_TYPE_OUTPUT)
		start = " >";
	else if (redirect_type == REDIRECT_TYPE_INPUT)
		start = "<";
	else if (redirect_type == REDIRECT_TYPE_ERROR)
		start = "2>";
	else
		start = ">";
	output_p = strstr(command_line, start);
	first = output_p;
	d = dest;
	if (output_p == NULL) {
		*dest = 0;
		return NULL;
	}
	output_p += strlen(start);
	if (redirect_type != REDIRECT_TYPE_INPUT) {
		if (*output_p == '>') {
			if (is_append != NULL)
				*is_append = 1;
			output_p++;
		}
		else {
			if (is_append != NULL)
				*is_append = 0;
		}
	}
	while (isspace(*output_p) && (*output_p != 0))
		output_p++;
	while (!isspace(*output_p) && (*output_p != 0)) {
		*d = *output_p;
		d++;
		output_p++;
	}
	*d = 0;
	return first;
}


/*
	get the complete argv matrix from the command line "line" and fills the output_redirect, output_append, error_redirect, error_append and input_redirect if necesary.
*/
char **get_complete_argv(char *line, char *output_redirect, int *output_append, char *error_redirect, int *error_append, char *input_redirect) {
	char *p;
	int p_len;
	int argc = 0;
	char **ret = NULL;
	char *temp_line = malloc(strlen(line) + 1);
	char *input_p;
	char *output_p;
	char *error_p;
	char *command_line_finish;

	input_p = get_redirect_name(line, REDIRECT_TYPE_INPUT, input_redirect, NULL);
	output_p = get_redirect_name(line, REDIRECT_TYPE_OUTPUT, output_redirect, output_append);
	error_p = get_redirect_name(line, REDIRECT_TYPE_ERROR, error_redirect, error_append);
	command_line_finish = input_p;
	if (output_p != NULL)
		command_line_finish = (output_p < command_line_finish) ? output_p : command_line_finish;
	if (error_p != NULL)
		command_line_finish = (error_p < command_line_finish) ? error_p : command_line_finish;
	if (command_line_finish != NULL)
		*command_line_finish = 0;
	if ((input_p != NULL) || (output_p != NULL) || (error_p != NULL)) {
		if (input_p == NULL)
			command_line_finish = output_p;
		else if (output_p == NULL)
			command_line_finish = input_p;
		else
			command_line_finish = (input_p < output_p) ? input_p : output_p;
		*command_line_finish = 0;
	}
	strcpy(temp_line, line);
	while ((p = strsep(&temp_line, " ")) != NULL) {
		if (strcmp(p, "") == 0)
			continue;
		p_len = strlen(p);
		ret = realloc(ret, (argc + 1) * sizeof(char *));
		ret[argc] = malloc(p_len + 1);
		memcpy(ret[argc], p, p_len + 1);
		argc++;
		/*printf("\t\t'%s'\n", p);*/
	}
	ret = realloc(ret, (argc + 1) * sizeof(char *));
	ret[argc] = NULL;
	argc++;
	free(temp_line);
	return ret;
}


/*
	free the argv matrix
*/
void free_complete_argv(char **argv) {
	char **p = argv;

	while (*p != NULL) {
		/*printf("\t\t\t...'%s'\n", *p);*/
		free(*p);
		*p++;
	}
	free(argv);
}


/*
	empty signal handler for SIGINT signal (Ctrl-C)
*/
void signal_handler(int signal) {
	/*
	fprintf(stderr, "SIGNAL %d received by %d (ppid=%d child_pid=%d)\n", signal, getpid(), getppid(), child_pid);
	if (child_pid != 0) {
		kill(child_pid, SIGINT);
		printf("-- CTRL-C (kill child) --\n");
	}
	else
		fprintf(stderr, "ush: No child process to kill\n");
	*/
}


/*
	run the specified executable "file" with the "argv" argument array. If specified, the stdin, stdout and strerr are redirected.
*/
void run(char *file, char **argv, char *output_redirect, int output_append, char *error_redirect, int error_append, char *input_redirect) {
	/*printf("ush: '%s' found\n", complete_path);*/
	child_pid = fork();
	if (child_pid == 0) {
		int output_fd = -1;
		int input_fd = -1;
		int error_fd = -1;
		/* child */
		if ((output_redirect != NULL) && (*output_redirect != 0)) {
			/* redirect output */
			int flags = O_WRONLY | O_CREAT | ((output_append) ? O_APPEND : O_TRUNC);
			if ((output_fd = open(output_redirect, flags, 0644)) < 0) {
				fprintf(stderr, "ush: unable to redirect output to '%s'\n", output_redirect);
				return;
			}
		}
		if ((error_redirect != NULL) && (*error_redirect != 0)) {
			/* redirect error */
			int flags = O_WRONLY | O_CREAT | ((error_append) ? O_APPEND : O_TRUNC);
			if ((error_fd = open(error_redirect, flags, 0644)) < 0) {
				fprintf(stderr, "ush: unable to redirect error to '%s'\n", error_redirect);
				return;
			}
		}
		if ((input_redirect != NULL) && (*input_redirect != 0)) {
			/* redirect input */
			if ((input_fd = open(input_redirect, O_RDONLY)) < 0) {
				fprintf(stderr, "ush: unable to redirect input from '%s'\n", input_redirect);
				return;
			}
		}
		if (input_fd != -1)
			dup2(input_fd, 0);
		if (output_fd != -1)
			dup2(output_fd, 1);
		if (error_fd != -1)
			dup2(error_fd, 2);
		execvp(file, argv);
		fprintf(stderr, "ush: unable to execute '%s'\n", file);
		exit(-1);
	}
	else if (child_pid < 0) {
		/* error */
		fprintf(stderr, "ush: unable to fork '%s'\n", file);
	}
	else {
		/* parent */
		wait(NULL);
		child_pid = 0;
	}
}


int main(int argc, char **argv, char **envp) {
	/* install a dummy signal handler for SIGINT (Ctrl-C) */
	prev_signal_handler = signal(SIGINT, signal_handler);
	/* if no arguments passed, by default get input from stdin */
	input_file = stdin;
	if (argc > 1)
		input_file = fopen(argv[1], "rt");
	while (1) {
		/* print the prompt only if input is stdin */
		if (input_file == stdin) {
			printf("ush-0.1$ ");
			fgets(input, 1024, input_file);
		}
		else {
			if (fgets(input, 1024, input_file) == NULL)
				exit(0);
		}
		/* replace las character (\n) by 0 */
		input[strlen(input) - 1] = 0;
		if (strlen(input) < 1)
			continue;
		/* expand the envorinment variables at command line ("expand" function writes to "expanded" variable) */
		expand(input);
		memcpy(input, expanded, 1024);
		if (input[0] == '#') {
			/* comment */
			continue;
		}
		else if (strcmp(input, "exit") == 0) {
			/* exit from shell */
			return 0;
		}
		else if (strncmp(input, "cd ", 3) == 0) {
			/* change dir */
			if (chdir(&input[3]) != 0) {
				fprintf(stderr, "ush: cannot change dir\n");
			}
		}
		else if (strncmp(input, "export ", 7) == 0) {
			/* set variable */
			char *p = &input[7];
			char *separator = NULL;
			separator = strchr(p, '=');
			if (separator == NULL) {
				fprintf(stderr, "ush: syntax error: usage: export VAR=value\n");
			}
			else {
				char *variable_name = p;
				char *variable_value = separator + 1;
				*separator = 0;
				setenv(variable_name, variable_value, 1);
			}
		}
		else {
			/* try to execute command */
			char **new_argv = get_complete_argv(input, output_redirect_file, &output_redirect_append, error_redirect_file, &error_redirect_append, input_redirect_file);
			/*
			printf("ush: output_redirect = '%s' (append=%d)\n", output_redirect_file, output_redirect_append);
			printf("ush: error_redirect = '%s' (append=%d)\n", error_redirect_file, error_redirect_append);
			printf("ush: input_redirect  = '%s'\n", input_redirect_file);
			*/
			run(new_argv[0], new_argv, output_redirect_file, output_redirect_append, error_redirect_file, error_redirect_append, input_redirect_file);
			free_complete_argv(new_argv);
		}
	}
	return 0;
}
